from .utilitaires import entier_en_indice_unicode, tableau_en_array_latex
from typing_extensions import Self

def chiffres_base(base: int) -> list[str]:
    """Renvoie la liste des chiffres de la base (entier >= 2)."""
    assert isinstance(base, int), "base doit être du type int !"
    assert base >= 2, "base doit être supérieur ou égal à 2 !"
    res = []
    i = 0
    while i < base:
        if i < 10:
            res.append(str(i))
        else:
            res.append(chr(ord('A') + i - 10))
        i += 1
    return res

def zero(base: int):
    return EntierNat('0', base)

class EntierNat:

    def __init__(self, chaine: str, base: int = 10) -> None:
        assert isinstance(chaine, str), "chaine doit être du type str !"
        assert isinstance(base, int), "base doit être du type int !"
        assert base >= 2, "base doit être supérieur ou égal à 2 !"
        self.chaine = chaine
        self.base = base
        self.chiffres_base = chiffres_base(self.base)
        assert self._est_valide(), "Nombre invalide : chiffres et base incompatibles !"
        self.decimal = self._decimal()
        self.latex = self._latex()

    def _est_valide(self) -> bool:
        """Vérifie que les caractères utilisés dans la chaîne sont compatibles
        avec la base choisie."""
        for car in self.chaine:
            if car not in self.chiffres_base:
                return False
        return True
    
    def _decimal(self) -> int:
        """Renvoie la valeur décimale du nombre."""
        res = 0
        puissance = len(self.chaine) - 1
        for chiffre in self.chaine:
            val_chiffre = self.chiffres_base.index(chiffre)
            res += val_chiffre * self.base ** puissance
            puissance -= 1
        return res
    
    def changer_base(self, nouvelle_base: int, nb_chiffres_min: int = 1) -> None:
        """Change la base du nombre. Complète éventuellement le nombre par des
        zéros par la gauche si la chaîne du nombre après le changement de base
        est plus petit que nb_chiffres_min."""
        # Changement de la base
        self.base = nouvelle_base
        # Chiffres de la nouvelle base
        self.chiffres_base = chiffres_base(self.base)
        # Divisions successives
        chaine = ''
        quotient = self.decimal
        while quotient > 0:
            reste = quotient % self.base
            chaine = self.chiffres_base[reste] + chaine
            quotient = quotient // self.base
        if nb_chiffres_min > len(chaine):
            nb_zeros_a_ajouter = nb_chiffres_min - len(chaine)
            chaine = nb_zeros_a_ajouter * '0' + chaine
        self.chaine = chaine
        self.latex = self._latex()

    def _latex(self) -> str:
        """Renvoie une chaîne contenant le code LaTeX du nombre."""
        return r'(\text{' + self.chaine + '})_{' + str(self.base) + '}'

    def _repr_latex_(self) -> str:
        """Méthode spéciale spécifique à sympy qui permet de personnaliser
        l'affichage obtenu en dernière ligne de cellule dans un cahier Jupyter.
        """
        return self.latex
        
    def __str__(self):
        return '(' + self.chaine + ')' + entier_en_indice_unicode(self.base)
    
    def __repr__(self):
        return self.__str__()
    
    def __add__(self, autre: Self) -> Self:
        somme = self.decimal + autre.decimal
        res = EntierNat(str(somme), 10)
        res.changer_base(self.base)
        return res
    
    def __mul__(self, autre: Self) -> Self:
        somme = self.decimal * autre.decimal
        res = EntierNat(str(somme), 10)
        res.changer_base(self.base)
        return res
    
    def __eq__(self, autre: Self) -> bool:
        return self.decimal == autre.decimal
    
    def chiffres(self) -> list[Self]:
        """Renvoie le tableau des chiffres du nombre."""
        return [EntierNat(self.chaine[i], self.base) for i in range(len(self.chaine))]
    
    def nbre_chiffres(self) -> int:
        return len(self.chaine)


class TabConvChiffres:

    def __init__(self, base_depart: int, base_arrivee: int) -> None:
        assert isinstance(base_depart, int), "base_depart doit être un entier !"
        assert isinstance(base_arrivee, int), "base_arrivee doit être un entier !"
        assert base_depart >= base_arrivee, "base_depart doit être strictement superieur à base_arrivee !"
        self.base_depart = base_depart
        self.base_arrivee = base_arrivee
        self.tab = self._creer_tab()
        self.latex = tableau_en_array_latex(self.tab)   

    def _creer_tab(self) -> list[list[str]]:
        """Renvoie le tableau de conversion."""
        res = [
            [r'\text{écriture en base ' + str(self.base_depart) + '}'],
            [r'\text{écriture en base ' + str(self.base_arrivee) + '}']
        ]
        chiffres_base_depart = chiffres_base(self.base_depart)
        chiffre_max = chiffres_base_depart[-1]
        nbre_max = EntierNat(chiffre_max, self.base_depart)
        nbre_max.changer_base(self.base_arrivee)
        nb_chiffres_arrivee = len(nbre_max.chaine)
        for chiffre in chiffres_base_depart:
            nbre = EntierNat(chiffre, self.base_depart)
            res[0].append(nbre.latex)
            nbre.changer_base(self.base_arrivee, nb_chiffres_arrivee)
            res[1].append(nbre.latex)
        return res
    
    def __str__(self) -> str:
        return self.latex
    
    def __repr__(self) -> str:
        return self.latex
    
    def _repr_latex_(self) -> str:
        return self.latex
    

class TabOperation:

    def __init__(self, base: int, facteurs: list[str] = [], multiplication: bool = True) -> None:
        self.multiplication = multiplication
        assert isinstance(base, int), "base doit être un entier !"
        assert base >= 2, "base doit être supérieure ou égale à 2 !"
        self.base = base
        self.chiffres_base = chiffres_base(base)
        assert isinstance(facteurs, list), "facteurs doit être du type list !"
        for car in facteurs:
            assert isinstance(car, str), "facteurs ne doit contenir que des str !"
            assert car in self.chiffres_base, "facteurs ne doit contenir que des chiffres de la base !"
        self.facteurs = facteurs
        self.tab = self._creer_tab()
        self.latex = self._creer_latex()

    def _creer_tab(self):
        if self.multiplication:
            coin = r'\times'
        else: # addition
            coin = '+'
        res = [[coin] + [EntierNat(chiff, self.base).latex for chiff in self.chiffres_base]]
        if self.facteurs == []:
            self.facteurs = self.chiffres_base
        for fact in self.facteurs:
            facteur = EntierNat(fact, self.base)
            ligne = [facteur.latex]
            for chiff in self.chiffres_base:
                chiffre = EntierNat(chiff, self.base)
                if self.multiplication:
                    nbre = facteur * chiffre
                else: # addition
                    nbre = facteur + chiffre
                ligne.append(nbre.latex)
            res.append(ligne)
        return res
    
    def _creer_latex(self) -> str:
        format_col = 'c|' * len(self.tab[0])
        hlines = list(range(1, len(self.tab) + 1))
        return tableau_en_array_latex(self.tab, format_col, hlines)
    
    def __str__(self) -> str:
        return self.latex
    
    def __repr__(self) -> str:
        return self.latex
    
    def _repr_latex_(self) -> str:
        return self.latex


class AdditionPosee:

    def __init__(self, nbre1: str, nbre2: str, base: int) -> None:
        assert isinstance(base, int), "base doit être du type int !"
        assert base >= 2, "base doit être supérieur ou égal à 2 !"
        self.base = base
        self.n1 = EntierNat(nbre1, self.base)
        self.n2 = EntierNat(nbre2, self.base)
        if self.n1.decimal < self.n2.decimal:
            tmp = self.n1
            self.n1 = self.n2
            self.n2 = tmp
        self.tab = self._creer_tab()
        self.latex = self._creer_latex()

    def _creer_tab(self) -> list[EntierNat]:
        chiffres1 = [zero(self.base)] + self.n1.chiffres()
        retenues = [zero(self.base)] * len(chiffres1)
        chiffres2 = [zero(self.base)] * (len(chiffres1) - self.n2.nbre_chiffres()) + self.n2.chiffres()
        sommes = [zero(self.base)] * len(chiffres1)
        for i in range(len(chiffres2) + 1):
            res = retenues[-i] + chiffres1[-i] + chiffres2[-i]
            res.changer_base(self.base, nb_chiffres_min=2)
            sommes[-i] = EntierNat(res.chaine[1], self.base)
            if i < len(chiffres2):
                retenues[-i-1] = EntierNat(res.chaine[0], self.base)
        if sommes[0] == zero(self.base):
            retenues = retenues[1:]
            chiffres1 = chiffres1[1:]
            chiffres2 = chiffres2[1:]
            sommes = sommes[1:]
        # retrait des zéros inutiles
        for j in range(len(chiffres1) - self.n1.nbre_chiffres()):
            if chiffres1[j] == zero(self.base):
                chiffres1[j] = ''
        for j in range(len(chiffres2) - self.n2.nbre_chiffres()):
            if chiffres2[j] == zero(self.base):
                chiffres2[j] = ''
        for j in range(len(retenues)):
            if retenues[j] == zero(self.base):
                retenues[j] = ''
        tab = [
            ['', ''] + retenues + [''],
            ['', '('] + chiffres1 + [')_{' + str(self.base) + '}'],
            ['+', '('] + chiffres2 + [')_{' + str(self.base) + '}'],
            ['', '('] + sommes + [')_{' + str(self.base) + '}'],
        ]
        for i in range(len(tab)):
            for j in range(len(tab[0])):
                if isinstance(tab[i][j], EntierNat):
                    tab[i][j] = r'\text{' + tab[i][j].chaine + '}'
        for j in range(len(tab[0])):
            # retenues
            if tab[0][j] == '0' or tab[0][j] == '':
                tab[0][j] = ''
            else:
                tab[0][j] = '_' + tab[0][j]
            # chiffres1
            if j < len(tab[0]) - self.n1.nbre_chiffres() - 1 and tab[1][j] == '0':
                 tab[1][j] = ''
            # chiffres2
            if j < len(tab[0]) - self.n2.nbre_chiffres() - 1 and tab[2][j] == '0':
                 tab[2][j] = ''
        # sommes
        if tab[3][2] == '0' and len(tab[0]) > 4:
            tab[3][2] = ''
        return tab
    
    def _creer_latex(self) -> str:
        format_col = 'c' * len(self.tab[0])
        hlines = [len(self.tab) - 1]
        return tableau_en_array_latex(self.tab, format_col, hlines)
    
    def __str__(self) -> str:
        return self.latex
    
    def __repr__(self) -> str:
        return self.latex
    
    def _repr_latex_(self) -> str:
        return self.latex


class MultiplicationPosee:

    def __init__(self, nbre1: str, nbre2: str, base: int) -> None:
        assert isinstance(base, int), "base doit être du type int !"
        assert base >= 2, "base doit être supérieur ou égal à 2 !"
        self.base = base
        self.n1 = EntierNat(nbre1, self.base)
        self.n2 = EntierNat(nbre2, self.base)
        if self.n1.decimal < self.n2.decimal:
            tmp = self.n1
            self.n1 = self.n2
            self.n2 = tmp
        self.tab = self._creer_tab()
        self.latex = self._creer_latex()

    def _creer_tab(self) -> list[EntierNat]:
        nchiffres1 = self.n1.nbre_chiffres()
        nchiffres2 = self.n2.nbre_chiffres()
        produit = self.n1 * self.n2
        if nchiffres2 == 1:
            nlig = 3
        else:
            nlig = 3 + nchiffres2
        ncol = max(nchiffres1 + 2, produit.nbre_chiffres()) + 1
        tab = [['' for _ in range(ncol)] for _ in range(nlig)]
        # Nombre avant le trait d'opération
        for i in range(nchiffres1):
            tab[0][-i-1] = self.n1.chiffres()[-i-1]
        for i in range(nchiffres2):
            tab[1][-i-1] = self.n2.chiffres()[-i-1]
        # Ajout du symbole fois
        tab[1][-nchiffres1-2] = r'\times'
        # Calculs intermédiaires
        if self.n2.nbre_chiffres() > 1:
            for i in range(nchiffres2):
                prod = tab[1][-i-1] * self.n1
                nchiffres = prod.nbre_chiffres()
                for j in range(nchiffres + i):
                    if j < i:
                        # ajout des points
                        tab[i + 2][-j-1] = r'\cdot'
                    else:
                        # ajout du produit
                        tab[i + 2][-j-1] = prod.chiffres()[-j-1+i]
        # ajout du résultat
        res = self.n1 * self.n2
        for i in range(res.nbre_chiffres()):
            tab[-1][-i-1] = res.chiffres()[-i-1]
        # Conversion des EntierNat en chaînes de caractères
        for i in range(nlig):
            for j in range(ncol):
                if isinstance(tab[i][j], EntierNat):
                    tab[i][j] = r'\text{' + tab[i][j].chaine + '}'
        # Ajout des parenthèses et indices
        tab[0][-nchiffres1-1] = '('
        tab[1][-nchiffres1-1] = '('
        for i in range(2, nlig):
            tab[i][0] = '('
        for i in range(nlig):
            tab[i].append(')_{' + str(self.base) + '}')
        return tab
    
    def _creer_latex(self) -> str:
        format_col = 'c' * len(self.tab[0])
        if self.n2.nbre_chiffres() == 1:
            hlines = [2]
        else:
            hlines = [2, len(self.tab) - 1]
        return tableau_en_array_latex(self.tab, format_col, hlines)

    def __str__(self) -> str:
        return self.latex
    
    def __repr__(self) -> str:
        return self.latex
    
    def _repr_latex_(self) -> str:
        return self.latex


if __name__ == '__main__':
    mul1 = MultiplicationPosee('1456', '467', 10)
    print(mul1)


from .utilitaires import entier_en_indice_unicode, tableau_en_array_latex
from .divisibilite import pgcd
from typing_extensions import Self


class ClasseCongruence:

    def __init__(self, nombre: int, module: int) -> None:
        assert isinstance(nombre, int), "nombre doit être un entier !"
        assert isinstance(module, int), "module doit être un entier !"
        assert module >= 2, "module doit être supérieur ou égal à 2 !"
        self.reste = nombre % module
        self.restes = [nbre for nbre in range(module)]
        self.module = module
        self.latex = self._creer_latex()

    def __eq__(self, autre) -> bool:
        return self.module == autre.module and self.reste == autre.reste

    def __add__(self, autre: Self) -> Self:
        assert self.module == autre.module, "Addition impossible : les classes de congruence n'ont pas le même module !"
        return ClasseCongruence(self.reste + autre.reste, self.module)

    def __mul__(self, autre: Self) -> Self:
        assert self.module == autre.module, "Multiplication impossible : les classes de congruence n'ont pas le même module !"
        return ClasseCongruence(self.reste * autre.reste, self.module)
    
    def __rtruediv__(self, autre: Self) -> Self:
        return ClasseCongruence(autre, self.module) * self.inv()
    
    def __pow__(self, puissance: int) -> Self:
        assert isinstance(puissance, int), "puissance doit être un entier !"
        if puissance < 0:
            facteur = self.inv()
            puissance = abs(puissance)
        else:
            facteur = self
        res = ClasseCongruence(1, self.module)
        for _ in range(puissance):
            res *= facteur
        return res

    def __str__(self) -> str:
        return '[' + str(self.reste) + ']' + entier_en_indice_unicode(self.module)
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def _repr_latex_(self) -> str:
        return self.latex

    def _creer_latex(self):
        return '[' + str(self.reste) + ']_{' + str(self.module) + '}'

    def est_inversible(self) -> bool:
        return pgcd(self.reste, self.module) == 1
    
    def inv(self, erreur=True) -> Self|None:
        assert self.est_inversible(), f"Pas d'inverse pour {self.__str__()} !"
        for reste in self.restes:
            if reste * self.reste % self.module == 1:
                return ClasseCongruence(reste, self.module)
           

class TabOperationMod:

    def __init__(self, module: int, multiplication: bool = False) -> None:
        assert isinstance(module, int), "module doit être un entier !"
        assert module >= 2, "module doit être supérieur ou égal à 2 !"
        self.module = module
        self.restes = ClasseCongruence(0, module).restes
        self.classes = [ClasseCongruence(nbre, self.module) for nbre in self.restes]
        self.multiplication = multiplication
        self.tab = self._creer_tab()
        self.latex = self._creer_latex()

    def __str__(self) -> str:
        return self.latex
    
    def __repr__(self) -> str:
        return self.latex
    
    def _repr_latex_(self) -> str:
        return self.latex

    def _creer_latex(self) -> str:
        format_col = 'c|' * len(self.tab[0])
        hlines = list(range(1, len(self.tab) + 1))
        return tableau_en_array_latex(self.tab, format_col, hlines)

    def _creer_tab(self) -> list[str]:
        if self.multiplication:
            coin = r'\times'
        else: # addition
            coin = '+'
        res = [[coin] + [classe.latex for classe in self.classes]]
        for classe1 in self.classes:
            ligne = [classe1.latex]
            for classe2 in self.classes:
                if self.multiplication:
                    resultat = classe1 * classe2
                else: # addition
                    resultat = classe1 + classe2
                ligne.append(resultat.latex)
            res.append(ligne)
        return res


if __name__ == "__main__":
    cong1 = ClasseCongruence(7, 10)
    cong2 = ClasseCongruence(2, 10)
    print(cong1.est_inversible())
    print(cong2.est_inversible())
    print(cong1.inv())
    print(cong1**-1)
    print(1/cong1)


from .utilitaires import preparer_coef, tableau_en_array_latex, envelopper_latex_dans_balises_align, execute_dans_google_colab

def pgcd(a: int, b: int) -> int:
    assert isinstance(a, int), "a doit être un entier !"
    assert isinstance(b, int), "b doit être un entier !"
    assert a != 0 or b != 0, "0 et 0 n'ont pas de PGCD !"
    a = abs(a)
    b = abs(b)
    if a < b:
        a, b = b, a
    if b == 0:
        return a
    while a % b != 0:
        tmp = b
        b = a % b
        a = tmp
    return b


class EuclideEtendu:

    def __init__(self, nbre1: int, nbre2: int) -> None:
        assert isinstance(nbre1, int), "nbre1 doit être du type int !"
        assert isinstance(nbre2, int), "nbre2 doit être du type int !"
        assert nbre1 != 0, "nbre1 doit être non-nul !"
        assert nbre2 != 0, "nbre2 doit être non-nul !"
        if abs(nbre2) > abs(nbre1):
            tmp = nbre1
            nbre1 = nbre2
            nbre2 = tmp
        self.nbre1 = nbre1
        self.nbre2 = nbre2
        self.tab = self._creer_tab()
        self.equation = self._creer_equation()
        self.latex = self._creer_latex()
        
    def _creer_tab(self) -> list[str]:
        reste = abs(self.nbre1) % abs(self.nbre2)
        tab = [
            [r'\text{ligne}', 'x', 'y', 'ax + by \\text{ (reste)}', r'\text{quotient}'],
            [        'L_{1}',   1,   0,            abs(self.nbre1),                 ''],
            [        'L_{2}',   0,   1,            abs(self.nbre2),                 '']
        ]
        i = 3
        while reste != 0:
            quotient = tab[-2][3] // tab[-1][3]
            tab[-1][4] = quotient
            reste = tab[-2][3] % tab[-1][3]
            tab.append([''] * 5)
            tab[-1][3] = reste
            if reste != 0:
                tab[-1][0] = 'L_{' + str(i) + '} = L_{' + str(i - 2) + '}' + preparer_coef(-quotient) + 'L_{' + str(i - 1) + '}'
                tab[-1][1] = tab[-3][1] - quotient * tab[-2][1]
                tab[-1][2] = tab[-3][2] - quotient * tab[-2][2]
            i += 1
        for i in range(len(tab)):
            for j in range(len(tab[0])):
                if not isinstance(tab[i][j], str):
                    tab[i][j] = str(tab[i][j]) 
        return tab
    
    def _creer_equation(self):
        x = int(self.tab[-2][1])
        y = int(self.tab[-2][2])
        dernier_reste = str(self.tab[-2][3])
        if self.nbre1 < 0:
            x = -x
        if self.nbre2 < 0:
            y = -y
        xbold = r'\boldsymbol{' + str(x) + '}'
        if y < 0:
            ybold = r'\boldsymbol{(' + str(y) + ')}'
        else:
            ybold = r'\boldsymbol{' + str(y) + '}'
        nbre1 = str(self.nbre1)
        nbre2 = str(self.nbre2)
        if self.nbre1 < 0:
            nbre1 = '(' + nbre1 + ')'
        if self.nbre2 < 0:
            nbre2 = '(' + nbre2 + ')'
        membre_gauche_normal = nbre1 + r' \wedge ' + nbre2
        membre_gauche_positif = str(abs(self.nbre1)) + r' \wedge ' + str(abs(self.nbre2))
        membre_droite = xbold + r' \times ' + nbre1 + ' + ' + ybold + r' \times ' + nbre2
        if self.nbre1 > 0 and self.nbre2 > 0:
            return membre_gauche_normal + ' = ' + dernier_reste + ' = ' + membre_droite + '\n'
        else:
            return membre_gauche_normal + ' = ' + membre_gauche_positif + ' = ' + dernier_reste + ' = ' + membre_droite + '\n'
    
    def _creer_latex(self) -> str:
        format_col = 'l|ccccc'
        hlines = [1]
        res = tableau_en_array_latex(self.tab, format_col, hlines)
        res += r'\\' + '\n' + self.equation
        if not execute_dans_google_colab():
            return envelopper_latex_dans_balises_align(res)
        else:
            return res
    
    def __str__(self) -> str:
        return self.latex
    
    def __repr__(self) -> str:
        return self.latex
    
    def _repr_latex_(self) -> str:
        return self.latex


if __name__ == '__main__':
    print(pgcd(0, 0))


from sympy import Number, Rational, latex
from os import getenv

def preparer_coef(
        coef: int|float|Rational|Number,
        en_tete: bool = False,
        bold: bool = False
    ) -> str:
    r"""Renvoie dans une chaîne au format LaTeX un coefficient coef (type
    nombre) à écrire à gauche d'une variable dans une expression LaTeX. Si
    le coefficient est en tête de formule, choisir l'option en_tete=True.
    Exemples :
    >>> preparer_coef(-3)
    ' - 3'
    >>> preparer_coef(-3, en_tete=True)
    '-3'
    >>> preparer_coef(1)
    ' + '
    >>> preparer_coef(1, en_tete=True)
    ''
    >>> preparer_coef(Rational(-2, 3))
    r' - \frac{2}{3}'
    """
    if coef == 0:
        return ''
    elif coef == 1:
        if en_tete:
            return ''
        else:
            return ' + '
    elif coef == -1:
        if en_tete:
            return '-'
        else:
            return ' - '
    elif coef > 0:
        if en_tete:
            if bold:
                return r'\boldsymbol{' + latex(coef) + '}'
            else:
                return latex(coef)
        else:
            if bold:
                return ' + ' + r'\boldsymbol{' + latex(coef) + '}'
            else:
                return ' + ' + latex(coef)
    else: # coef < 0
        if en_tete:
            if bold:
                return r'\boldsymbol{' + '-' + latex(abs(coef)) + '}'
            else:
                return '-' + latex(abs(coef))
        else:
            if bold:
                return ' ' + r'\boldsymbol{-}' + ' ' + r'\boldsymbol{' + latex(abs(coef)) + '}'
            else:
                return ' - ' + latex(abs(coef))


def entier_en_indice_unicode(entier: int) -> str:
    """Renvoie une chaîne de caractère contenant le nombre entier positif passé
    en paramètre sous forme de petits chiffres de catalogue Unicode."""
    res = ''
    for chiffre in str(entier):
        res += chr(ord('₀') + ord(chiffre) - ord('0'))
    return res

def tableau_en_array_latex(tab: list[list[str]], format_col: str = '', hlines: list[int] = []) -> str:
    """Prend en paramètre un tableau 2D de chaîne de caractères et renvoie le
    code LaTeX de l'array correspondant. Attention, chaque cellule doit
    contenir une expression mathématique valide. Pour passer du texte écrire
    dans une case \text{<votre texte>}. Par défaut, les cellules sont centrées
    horizontalement et toutes les bordures des cellules sont tracées. On peut
    préciser le format des colonnes (ex: '|c|llll|r') et le numéro des lignes
    avant lesquelles tracer une hline."""
    assert isinstance(tab, list), "tab doit être du type list !"
    for ligne in tab:
        assert isinstance(ligne, list), "Chaque élément de tab doit être du type list !"
        for case in ligne:
            assert isinstance(case, str), "Chaque case de tab doit être du type str !"
    largeur = len(tab[0])
    for ligne in tab:
        assert len(ligne) == largeur, "Les lignes du tableau n'ont pas toutes le même nombre de cases !"
    assert isinstance(hlines, list), "hlines doit être du type list !"
    if format_col == '':
        format_col = largeur * '|c' + '|'
    for elt in hlines:
        assert isinstance(elt, int), "hlines ne doit contenir que des entiers !"
        assert elt in range(len(tab) + 1), "hlines ne doit contenir que des indices positifs valides !"
    res = r'\begin{array}{' + format_col + '}\n'
    for i in range(len(tab)):
        if i in hlines or hlines == []:
            res += r'    \hline' + '\n'
        res += '    ' + ' & '.join(tab[i]) + r' \\' + '\n'
    if i + 1 in hlines or hlines == []:
        res += r'    \hline' + '\n'
    res += r'\end{array}' + '\n'
    return res


def envelopper_latex_dans_balises_align(latex: str) -> str:
    r"""Enveloppe le code LaTeX passé en paramètre entre des des balises
    \begin{align} et \end{align}. Cela doit être fait uniquement si le
    Jupyter Notebook n'est pas exécuté dans Google Colab. Dans les
    environnements Jupyter Notebooks "classiques" le rendu est
    incorrect si les balises \begin{align} et \end{align} sont omises
    alors que c'est le contraire dans Google Colab.
    """
    return "\\begin{align*}\n" + latex + "\\end{align*}\n"


def execute_dans_google_colab() -> bool:
    return getenv("COLAB_RELEASE_TAG")

if __name__ == '__main__':
    assert preparer_coef(-3) == ' - 3'
    assert preparer_coef(-3, en_tete=True) == '-3'
    assert preparer_coef(1) == ' + '
    assert preparer_coef(1, en_tete=True) == ''
    assert preparer_coef(Rational(-2, 3)) == r' - \frac{2}{3}'
    print(entier_en_indice_unicode(16))